/*
 * Copyright Contributors to the fusa-gcc-plugin project
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */
#pragma once

#include <optional>
#include <unordered_set>
#include <vector>

#include "gcc-plugin.h"
#include "input.h"

#include "cryptoutils.hpp"
#include "declhelpers.hpp"
#include "distribution.hpp"
#include "macrohelpers.hpp"
#include "fusa.hpp"

using namespace std;

enum api_type { API_TYPE_FUNCTION = 0, API_TYPE_MACRO };

struct candidate {
  string name;
  string decl_source_file;
  location_t source_location;
  location_t invocation_location;
  vector<string> namespace_hierarchy;
  vector<string> class_hierarchy;
  enum api_type api_type;

  struct hash {
    size_t operator()(const candidate &s) const {
      char *p = NULL;

      if (s.api_type == API_TYPE_MACRO) {
        (void)asprintf(&p, "%s:%s", s.name.c_str(), s.decl_source_file.c_str());
      } else {
        (void)asprintf(&p, "%s:%s:%u", s.name.c_str(),
                       s.decl_source_file.c_str(), s.source_location);
      }

      string st(p);
      free(p);

      return ::hash<string>()(st);
    }
  };

  typedef unordered_set<candidate, candidate::hash> set;

  struct collection {
    candidate::set *items;

    collection() {
      this->items = new candidate::set();
    }

    void qualify_tree_recursive(tree t, location_t loc) {
      vector<tree> decls = declhelpers::find_function_decls(t);

      for (tree &t : decls)
        qualify(t, loc);
    }

    void qualify(tree t, location_t l) {
      if (optional<candidate> c = candidate::from_decl(t, l))
        items->insert(*c);
    }
  };

  explicit candidate(const char *name, const char *decl_source_file,
                     location_t source_location, location_t invocation_location,
                     vector<string> ns, vector<string> cs,
                     enum api_type at = API_TYPE_FUNCTION) {
    this->name = name;
    this->decl_source_file = decl_source_file;
    this->source_location = source_location;
    this->invocation_location = invocation_location;
    this->class_hierarchy = cs;
    this->namespace_hierarchy = ns;
    this->api_type = at;
  }

  static optional<candidate> from_location(location_t loc) {
    if (optional<macrohelpers::macro> macro =
            macrohelpers::get_macro_expansion(loc)) {
      if (distribution::is_distribution_file(macro->file_name.c_str())) {
        /**
         * TODO: Currently both hash and operator== is made to ignore
         * source_location for macros so that we can easily just hash
         * the candidate without much additional shenanigans. The shenanigans:
         *
         *  1| #define MAC(x, y) (x << y)
         *  2|
         *  3| int main(int argc, char *argv[]) {
         *  4|   return MAC(1, argc);
         *  5| }
         *
         * Line (4) after macro expansion looks like this:
         *
         *  4|   return (1 << argc);
         *
         * Which is the form we're processing here. Now the caveat is that
         * there are 2 nodes, `1` and `argc`, which will bear the mark of
         * being result of a macro expansion, each with their distinct location.
         *
         * If we take source_location into account, we'll register two distinct
         * macro invocations because of the differing `column` value for each
         *node.
         *
         * The problem is merely cosmetic, the user receives error messages such
         *as:
         *
         *  4|   return MAC(1, argc);
         *   |                   ^
         *
         * Notice how the ^caret points "somewhere" in the middle of the
         * expansion site, where in fact we want it to point at the start
         * of the macro like this:
         *
         *  4|   return MAC(1, argc);
         *   |          ^
         *
         * Simple solution for this is to (probably) just store all the
         * possible expansion locations and pick the one with the lowest
         * LOCATION_COLUMN() value.
         **/
        return candidate(macro->name.c_str(), macro->file_name.c_str(),
                         macro->where, loc, {}, {}, API_TYPE_MACRO);
      }
    }

    return nullopt;
  }

  static optional<candidate> from_decl(tree decl, location_t location) {
    // Be lenient about decl
    if (!decl)
      return nullopt;

    // Attempt dereference an ADDR_EXPR, might be a FUNCTION_DECL
    if (TREE_CODE(decl) == ADDR_EXPR)
      decl = TREE_OPERAND(decl, 0);

    // If we're not dealing with a FUNCTION_DECL, skip over
    if (TREE_CODE(decl) != FUNCTION_DECL)
      return nullopt;

    const char *source_file = DECL_SOURCE_FILE(decl);
    // If we're not dealing with something owned by distribution, skip over
    if (!distribution::is_distribution_file(source_file))
      return nullopt;

    const char *name = DECL_NAME_STRING(decl);

    // If we're dealing with function pointers, skip over
    if (strncmp(name, "(nofn)", 6) == 0)
      return nullopt;

    // Skip items synthesized by the compiler (intlists, ctors, dtors...)
    if (DECL_ARTIFICIAL(decl))
      return nullopt;

    // const char *asm_name = IDENTIFIER_POINTER(DECL_ASSEMBLER_NAME(decl));
    int decl_location = DECL_SOURCE_LINE(decl);

    return candidate(name, source_file, decl_location, location,
                     declhelpers::get_namespace_hierarchy(decl),
                     declhelpers::get_class_hierarchy(decl));
  }

  bool operator==(const candidate &other) const;
  bool fusa_check(vector<certified_symbol> &certified_symbols, candidate::set &apis,
                  struct plugin_args &args) const;
};
